import argparse
from anytracker import Anytracker
from video import VideoCapture
import numpy as np
import pandas as pd
import cv2
import os.path as op
from tqdm import tqdm

def px(*args):
    return (int(round(arg)) for arg in args)

def main(input, nframes):
    ### create AnyTrack Tracking object
    track = Anytracker(input=input, output='output_anytrack')
    for video in track.videos:
        cap = VideoCapture(video, 0)
        w,h = cap.w, cap.h
        numframes=nframes
        writer = []
        for i, fly in enumerate(track.outdict['trajectory_files'][video]):
            _file = op.join(track.outdict['folders']['output'], 'overlay_{}_{}.avi'.format(op.basename(video).split('.')[0], i))
            print(_file)
            writer.append(cv2.VideoWriter(_file, cv2.VideoWriter_fourcc('M','J','P','G'), 30, (600, 600), True))
        for i in tqdm(range(numframes)):
            __, frame = cap.read()
            id = 0
            for fly, roi in zip(track.outdict['trajectory_files'][video], track.outdict['ROIs'][video]):
                data=pd.read_csv(fly)
                x,y=np.array(data['head_x']), np.array(data['head_y'])
                cx, cy, cr = px(roi['x'], roi['y'], 1.5*roi['radius'])
                output = frame.copy()
                output = output[cy-cr:cy+cr,cx-cr:cx+cr]
                #output = cv2.resize(output, (tuple(px(.5*w,.5*h))))
                cv2.circle(output, tuple(px(x[i]-(cx-cr), y[i]-(cy-cr))), 1, (255,0,255),-1)
                cv2.putText(output, str(id), tuple(px(x[i]-(cx-cr)+10, y[i]-(cy-cr)+10)), cv2.FONT_HERSHEY_SIMPLEX , .5, (255,0,255), 1, cv2.LINE_AA)

                writer[id].write(cv2.resize(output, (600,600)))
                id += 1
                #cv2.imshow("Overlay", output)
                #cv2.waitKey(0) # time to wait between frames, in mSec
        cap.release()
        for w in writer:
            w.release()

if __name__ == '__main__':
    ### arguments parsing
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', dest='input', action='store',
                        help='input file(s)/directory')
    parser.add_argument('-n', dest='nframes', action='store', type=int,
                        help='number of frames')
    args = parser.parse_args()
    input = args.input
    nframes = args.nframes
    main(input, nframes)



import cv2
import numpy as np

from anytrack.video import VideoCapture

def get_background(video, num_frames=30, how='uniform', offset=0, show_all=False, frames=None):
    print('Start modelling background...', flush=True)
    cap = VideoCapture(video,0)
    h, w = cap.h, cap.w
    bgframe = np.zeros((h, w), dtype=np.float32)
    if how=='uniform':
        if frames is None:
            choices = np.random.choice(cap.len-1, num_frames)
        else:
            choices = np.random.choice(frames, num_frames)
    for i in range(num_frames):
        if how=='uniform':
            frameno = choices[i] + offset
        else:
            frameno = i + offset
        frame = cap.get_frame(frameno)
        img = cv2.cvtColor(frame,cv2.COLOR_BGR2GRAY)
        bgframe += img/num_frames
    step = 1
    if show_all:
        cv2.imshow("BG model iteration {}".format(step), cv2.resize(bgframe.astype(np.uint8), (600,600)))
        cv2.waitKey(0) # time to wait between frames, in mSec
        cv2.destroyAllWindows()
    newbg = np.zeros((h, w), dtype=np.float64)
    bgcount = np.zeros((h, w), dtype=np.float64)
    bgcount[:] = 1.
    step += 1
    if how=='uniform':
        if frames is None:
            choices = np.random.choice(cap.len-1, num_frames)
        else:
            choices = np.random.choice(frames, num_frames)
    for i in range(num_frames):
        if how=='uniform':
            frameno = choices[i] + offset
        else:
            frameno = i + offset
        frame = cap.get_frame(frameno)
        difference = bgframe.astype(np.float64) - cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
        __, subtr = cv2.threshold(difference, 20, 255, cv2.THRESH_BINARY)
        bgmask = np.zeros((h, w), dtype=np.uint8)
        bgmask[subtr==0] = frame[subtr==0, 0]
        bgcount[subtr==0] += 1.
        newbg += bgmask.astype(np.float64)
    newbg = np.divide(newbg,bgcount)
    if show_all:
        cv2.imshow("BG model iteration {}".format(step), cv2.resize(newbg.astype(np.uint8), (600,600)))
        cv2.waitKey(0) # time to wait between frames, in mSec
        cv2.destroyAllWindows()
    return newbg



import numpy as np
import cv2

from anytrack.video import VideoCapture

def get_contours(frame, background, threshold_level=10, thresholding='dark'):
    bg = background.astype(np.float64)
    gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
    if thresholding == 'dark':
        difference =  bg - gray_frame
    elif thresholding == 'bright':
        difference =  gray_freame - bg
    else:
        difference =  np.abs(bg - gray_frame)
    __, subtr = cv2.threshold(difference, threshold_level, 255, cv2.THRESH_BINARY)
    contours, hierarchy = cv2.findContours(subtr.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    return contours

def get_contour_stats(video, background, force_num_contours=None, num_frames=50, how='uniform'):
    print('Infer contour statistics...', flush=True)
    cap = VideoCapture(video,0)
    h, w = cap.h, cap.w
    number_contours = np.zeros(num_frames)
    all_areas = []
    if how=='uniform':
        choices = np.random.choice(cap.len-1, num_frames)
    for i in range(num_frames):
        if how=='uniform':
            frameno = choices[i]
        else:
            frameno = i
        frame = cap.get_frame(frameno)
        contours = get_contours(frame, background)
        contours = [cnt for cnt in contours if cv2.contourArea(cnt) > 20]
        number_contours[i] = len(contours)
        for cnt in contours:
            all_areas.append(cv2.contourArea(cnt))
    mean_area = np.nanmean(all_areas)
    mean_number = np.nanmean(number_contours)
    if force_num_contours is None:
        area_contours = np.zeros((num_frames, int(round(mean_number))))
    else:
        area_contours = np.zeros((num_frames, force_num_contours))
    for i in range(num_frames):
        if how=='uniform':
            frameno = choices[i]
        else:
            frameno = i
        frame = cap.get_frame(frameno)
        contours = get_contours(frame, background)
        areas = sorted([cv2.contourArea(cnt) for cnt in contours], key=lambda x: np.abs(x-mean_area))[:int(round(mean_number))]
        if force_num_contours is None:
            area_contours[i, :] = np.array(areas[:int(round(mean_number))])
        else:
            area_contours[i, :] = np.array(areas[:force_num_contours])
    if force_num_contours is None:
        return mean_number, np.amin(area_contours), np.amax(area_contours)
    else:
        return force_num_contours, np.amin(area_contours), np.amax(area_contours)



import numpy as np
import pandas as pd

class Trajectory(object):
    def __init__(self, num_frames):
        self.columns = ['body_x', 'body_y', 'head_x', 'head_y', 'major', 'minor', 'angle', 'avg_px']
        self.all_columns = ['body_x', 'body_y', 'head_x', 'head_y', 'major', 'minor', 'angle', 'avg_px',
                            'ax', 'ay', 'ox', 'oy', 'apx', 'opx']
        self.data = np.zeros((num_frames,len(self.columns)+6))
        self.data[:] = np.nan
        self.x = self.data[:,0]
        self.y = self.data[:,1]
        self.hx = self.data[:,2]
        self.hy = self.data[:,3]
        self.major = self.data[:,4]
        self.minor = self.data[:,5]
        self.angle = self.data[:,6]
        self.i = 0

    def append(self, i, x=None, y=None, hx=None, hy=None, ma=None, mi=None, angle=None, avg_px=None, ax=None, ay=None, ox=None, oy=None, apx=None, opx=None):
        if x is not None:
            self.data[i, 0] = x
        if y is not None:
            self.data[i, 1] = y
        if hx is not None:
            self.data[i, 2] = hx
        if hy is not None:
            self.data[i, 3] = hy
        if ma is not None:
            self.data[i, 4] = ma
        if mi is not None:
            self.data[i, 5] = mi
        if angle is not None:
            self.data[i, 6] = angle
        if avg_px is not None:
            self.data[i, 7] = avg_px
        if ax is not None:
            self.data[i, 8] = ax
        if ay is not None:
            self.data[i, 9] = ay
        if ox is not None:
            self.data[i, 10] = ox
        if oy is not None:
            self.data[i, 11] = oy
        if apx is not None:
            self.data[i, 12] = apx
        if opx is not None:
            self.data[i, 13] = opx

    def save(self, _file):
        print('Saving flytracks to {}'.format(_file))
        pd.DataFrame(columns=self.columns, data=self.data[:,:len(self.columns)]).to_csv(_file, index_label='frame')

    def trace(self,i,_len):
        return self.data[i-_len:i,0:2]



import cv2
import numpy as np

class VideoCapture:
    def __init__(self, src, startframe):
        self.src = src
        self.cap = cv2.VideoCapture(self.src,0)
        self.start = startframe
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, startframe)
        self.grabbed, self.frame = self.cap.read()
        self.len = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.cap.set(cv2.CAP_PROP_POS_AVI_RATIO,1)
        self.duration = self.cap.get(cv2.CAP_PROP_POS_MSEC)/1000. ### in secs
        self.cap.set(cv2.CAP_PROP_POS_AVI_RATIO,0)
        self.w  = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    def get_average(self, nframes, random=False, to=None):
        avg = np.zeros(self.frame.shape, dtype=np.float32)
        if random:
            choices = np.random.choice(self.len-self.start, nframes)
        for i in range(nframes):
            if random:
                self.cap.set(cv2.CAP_PROP_POS_FRAMES, choices[i]+self.start)
            ret, frame = self.read()
            cv2.accumulate(frame,avg)
            img = cv2.convertScaleAbs(avg, alpha=1./nframes)
        if to is not None:
            img = cv2.resize(img, to)
        return img

    def get(self, var):
        return self.cap.get(var)

    def get_frame(self, var):
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, var)
        _, frame = self.read()
        return frame

    def set(self, var1, var2):
        self.cap.set(var1, var2)

    def set_frame(self, var):
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, var)

    def read(self):
        self.grabbed, self.frame = self.cap.read()
        frame = self.frame.copy()
        grabbed = self.grabbed
        return grabbed, frame

    def restart(self):
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, self.start)

    def release(self):
        self.cap.release()



import numpy as np

def rle(inarray):
    """ run length encoding. Partial credit to R rle function.
        Multi datatype arrays catered for including non Numpy
        returns: tuple (runlengths, startpositions, values) """
    ia = np.array(inarray, dtype=np.int32)                  # force numpy
    n = len(ia)
    if n == 0:
        return (None, None, None)
    else:
        y = np.array(ia[1:] != ia[:-1])     # pairwise unequal (string safe)
        i = np.append(np.where(y), n - 1)   # must include last element posi
        z = np.diff(np.append(-1, i))       # run lengths
        p = np.cumsum(np.append(0, z))[:-1] # positions
        return(z, p, ia[i]) # simply return array runlengths runlen, positions and states

def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """
    return np.isnan(y), lambda z: z.nonzero()[0]

def interpolate(arr):
    nans, x= nan_helper(arr)
    arr[nans]= np.interp(x(nans), x(~nans), arr[~nans])
    return arr

def smooth(y, box_pts):
    box = np.ones(box_pts)/box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth

def px(val):
    return int(round(val))

def dist(pos1, pos2):
    dx, dy = (pos1[0] - pos2[0]), (pos1[1] - pos2[1])
    return np.sqrt(dx * dx + dy * dy)

def in_roi(pos, roi):
    if dist(pos,(roi['x'], roi['y']))<=1.1*roi['outer']:
        return True
    else:
        return False




import cv2
import numpy as np
import math
import os, platform, subprocess
import os.path as op
import uuid
from pathlib import Path
import pprint as pp
import pytrack

from anytrack.video import VideoCapture

__author__ = "Dennis Goldschmidt"
__date__ = "29.08.2018"
__email__ = "dennis.goldschmidt@neuro.fchampalimaud.org"

### TODO: this should be in core.colors
GREEN = (139, 255, 45)
MAGENTA = (217, 28, 255)
YEAST = (0, 140, 255)
SUCROSE = (0, 140, 255)
MATCH = (51,255,255)

def px(val):
    return int(round(val))

class arenaROIselector(object):
    """ Class for selecting arena ROIs manually in a given videos.
        Returns dictionary of ROIs of the four arenas ('topleft', 'topright', 'bottomleft', 'bottomright')
        containing a dictionary with the following keys:
        -   x:      x coordinate of center position in pixel
        -   y:      y coordinate of center position in pixel
        -   radius: radius of arena circle in pixel
        -   angle:  orientation of first inner yeast spot within arena circle
    """
    def __init__(self, video, setup, windowName=None, nframes=300, random=True, pattern='6_6_radial', method='automated'):
        self.__scale = 1.6
        self.img, self.original_img = self.init_frame(video, setup, nframes, random)
        self.setup = setup
        self.__backup = self.img.copy()
        self.ROIs = []
        self.method = method
        self.selected = -1
        self.windowName = windowName
        self.__drawing = False
        self.__editing = False
        self.__automated_arenas = False
        self.loc, self.patches = [], []
        self.pattern = pattern

        if method == 'automated' or 'supervised':
            # automatic detection of arenas
            blur = cv2.GaussianBlur(self.original_img.copy()[:,:,0],(5,5),0)
            ret3,th3 = cv2.threshold(blur,0,255,cv2.THRESH_BINARY+cv2.THRESH_OTSU)
            contours, hierarchy = cv2.findContours(th3, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            self.auto_rois = contours
            for cnt in contours:
                if len(cnt) > 50:
                    (x,y),(ma,mi),angle = cv2.fitEllipse(cnt)
                    self.ROIs.append({'x': x, 'y': y, 'radius': ma/3, 'angle': 0})
            self.ROIs = sorted(self.ROIs, key=lambda k: (k['x'], k['y']))

        if method == 'supervised' or method == 'manual':
            if windowName != None:
                self.draw()
                cv2.setMouseCallback(self.windowName, self.click)
            else:
                self.windowName = "Arena ROI Selection"
                self.draw()
                cv2.setMouseCallback(self.windowName, self.click)
            self.mainloop()

    def init_frame(self, video, setup, nframes, random, start_frame=0):
        cap = VideoCapture(video,start_frame)
        ret, image = cap.read()
        self.__width, self.__height = int(image.shape[1]/self.__scale), int(image.shape[0]/self.__scale)
        image = cap.get_average(100, random=True, to=(self.__width, self.__height))
        original = image.copy()
        w, h = image.shape[0], image.shape[1]
        font = cv2.FONT_HERSHEY_SIMPLEX
        cv2.putText(image, 'Video: {}'.format(video), (20, h-35), font, .5, GREEN, 1, cv2.LINE_AA)
        cv2.putText(image, 'Setup: {}'.format(setup), (20, h-10), font, .5, GREEN, 1, cv2.LINE_AA)
        return image, original

    def mainloop(self):
        while True:
            k = cv2.waitKeyEx(1)
            if k == ord('r'):
                self.resetCanvas()
            elif k == ord('1'):
                self.select('topleft')
            elif k == ord('2'):
                self.select('topright')
            elif k == ord('3'):
                self.select('bottomleft')
            elif k == ord('4'):
                self.select('bottomright')
            elif k == 27:
                break
            elif k == 65535:
                self.delete()
            elif k == ord('c'):
                self.__automated_arenas = not self.__automated_arenas
                self.draw()
            elif k == 43 or k == ord('w'):
                self.scale_by(1)
            elif k == 45 or k == ord('s'):
                self.scale_by(-1)
            elif k == ord('d'):
                self.rotate_by(1)
            elif k == ord('a'):
                self.rotate_by(-1)
            elif k == 32:
                self.save_templates()
            elif k == 13:
                self.loc = self.match_templates('arena', self.setup)
                self.patches = self.get_peak_matches()
                self.draw()
            elif k == -1:
                continue
            else:
                print(k)

    def resetCanvas(self):
        """Function to reset the canvas with the given image. This resets the current ROI in memory, but leaves the entire ROI list untouched.
        Call this function from the mainloop selectively by listening to specific key strokes."""
        self.img = self.__backup.copy()
        self.ROIs = []
        self.selected = 'topleft'
        self.draw()


    def click(self, event, x, y, flags, param):
        """Main click event for the mouse. Allowed actions:
        Left click: If a ROI is open, that is, it is not enclosed, it adds another point, where the mouse clicked to the polygon
        Right click: If the ROI is open, then it closes the ROI polygon. This was done to make sure that the ROI is closed, since
                     even a pixel of open ROI, while invisible to the human eye, might wreak havoc for algorithms like flood-fill
                     and would need further prepocessing (Trust me, I faced it and this solves a bit of the headaches). Updates
                     the ROI list automatically when the polygon is closed. If the ROI list is empty, then you probably didn't close
                     the polygon. Try right clicking next time.
        Alternative Right click: This is triggered only when the polygon is closed. This starts the orientation mode, recognizable
                     by a line following the mouse from the centroid of the ROI. This mode is used to specify a guide from which
                     the orientation of the ROI is to be calculated.
        Alternative Left click: This is triggered only when the polygon is closed and the orientation mode has started. This
                     finalizes the orientation to face in the direction where the user clicks. So you get a line from which the
                     orientation of the ROI is estimated. The model assumes an upright, right handed, 360 degrees rotational frame.
                     """
        if event == cv2.EVENT_LBUTTONDBLCLK:
            for pt, w, v in self.patches:
                x0, y0 = pt
                if np.sqrt((x0-x)**2 + (y0-y)**2) < w:
                    self.ROIs[self.selected]['x'] = x0
                    self.ROIs[self.selected]['y'] = y0
                    self.ROIs[self.selected]['radius'] = w/2
                    self.ROIs[self.selected]['angle'] = 0
                    self.__editing = False
        elif event == cv2.EVENT_LBUTTONDOWN:
            self.__editing = False
            for id, arena in enumerate(self.ROIs):
                if arena is not None:
                    x0, y0 = arena['x'], arena['y']
                    if np.sqrt((x0-x)**2 + (y0-y)**2) < arena['radius']:
                        self.selected = id
                        self.__editing = True
                    elif self.selected == id:
                        self.next_selected()
            for pt, w, v in self.patches:
                x0, y0 = pt
                if np.sqrt((x0-x)**2 + (y0-y)**2) < w:
                    self.__editing = False
            if not self.__editing:
                self.__drawing = True
                self.x0, self.y0 = x, y
                self.ROIs.append({})
                self.ROIs[-1]['x'] = x
                self.ROIs[-1]['y'] = y
                self.ROIs[-1]['radius'] = 100
                self.ROIs[-1]['angle'] = 0
        elif event == cv2.EVENT_MOUSEMOVE:
            if self.__drawing:
                s = self.ROIs[self.selected]['radius']
                x0, y0 = self.ROIs[self.selected]['x'], self.ROIs[self.selected]['y']
                self.ROIs[self.selected]['radius'] = max(1, s+0.01*(x-x0))
                self.ROIs[self.selected]['angle'] = math.fmod(0.01*(y-y0), 2.*np.pi)
            elif self.__editing:
                self.ROIs[self.selected]['x'] = x
                self.ROIs[self.selected]['y'] = y
        elif event == cv2.EVENT_LBUTTONUP:
            if self.__drawing:
                self.__drawing = False
            elif self.__editing:
                self.__editing = False
        self.draw()

    def delete(self):
        self.ROIs.pop(self.selected)
        self.selected = len(self.ROIs)-1
        self.draw()

    def draw(self):
        img = self.img.copy()
        font = cv2.FONT_HERSHEY_SIMPLEX
        ### draw arenas
        for id, arena in enumerate(self.ROIs):
            if arena is not None:
                ### draw arenas
                x, y, r, a  = int(arena['x']), int(arena['y']), int(arena['radius']), arena['angle']
                if id == self.selected:
                    color = MAGENTA
                    lw = 2
                else:
                    color = GREEN
                    lw = 1
                cv2.circle(img,(x, y),r,color,lw)
                cv2.circle(img,(x, y),1,color,-1)
                cv2.rectangle(img, (px(x-1.5*r),px(y-1.5*r)), (px(x+1.5*r), px(y+1.5*r)), color, 1)
                ### draw spots
                if self.pattern == '6_6_radial':
                    for j in range(2):
                        for i in range(3):
                            sr = 1
                            if i==0 and j==0:
                                sr = 2
                            spot_r, angle = (j+1)*r/2.5, a+i*(2*np.pi/3)-j*(2*np.pi/12)
                            sx, sy = int(x + spot_r*np.cos(angle)), int(y + spot_r*np.sin(angle))
                            cv2.circle(img, (sx, sy), int(3*r/50),YEAST, sr)
                cv2.putText(img, str(id), (x+10, y-10), font, .5, color, 1, cv2.LINE_AA)
                cv2.putText(img, str(r), (x+10, y+10), font, .5, color, 1, cv2.LINE_AA)
        ### draw labels
        cv2.putText(img, 'Selected arena: {}'.format(self.selected), (20, 30), font, .75, (139, 255, 45), 1, cv2.LINE_AA)

        for pt, w, v in self.patches:
            cv2.rectangle(img, pt, (pt[0] + w, pt[1] + w), MATCH, 1)
            cv2.circle(img, (int(pt[0] + w/2), int(pt[1] + w/2)), int(w/2),MATCH, 1)

        if self.__automated_arenas:
            contours = self.auto_rois
            cv2.drawContours(img, contours, -1, (0,255,0), 3)
            for cnt in contours:
                if len(cnt) > 5:
                    ellipse = cv2.fitEllipse(cnt)
                    (x,y),(ma,mi),angle = ellipse
                    cv2.ellipse(img,ellipse,(0,255,0),2)
                    cv2.circle(img, (px(x), px(y)), px(ma/3), MATCH, 1)


        cv2.imshow(self.windowName, img)

    def next_selected(self):
        self.selected = len(self.ROIs)

    def get_spots(self, id):
        x, y, r, a = self.ROIs[id]['x']* self.__scale, self.ROIs[id]['y']* self.__scale, self.ROIs[id]['radius']* self.__scale, self.ROIs[id]['angle']
        a = a%(2.*np.pi/3)
        spots = []
        if self.pattern == '6_6_radial':
            for j in range(2):
                for i in range(3):
                    ### yeast spot
                    spot_id = 2*(i+j*3)
                    spot_r, angle = (j+1)*r/2.5, a+i*(2*np.pi/3)-j*(2*np.pi/12)
                    sx, sy = x + spot_r*np.cos(angle), y + spot_r*np.sin(angle)
                    spots.append({'id': spot_id, 'x': round(float(sx),1), 'y': round(float(sy),1), 'substrate': 'yeast'})
                    ### sucrose spot
                    spot_id = 2*(i+j*3)+1
                    spot_r, angle = (j+1)*r/2.5, a+i*(2*np.pi/3)-j*(2*np.pi/12)+np.pi/3
                    sx, sy = x + spot_r*np.cos(angle), y + spot_r*np.sin(angle)
                    spots.append({'id': spot_id, 'x': round(float(sx),1), 'y': round(float(sy),1), 'substrate': 'sucrose'})
        return spots

    def rotate_by(self, val):
        if self.ROIs[self.selected] is not None:
            self.ROIs[self.selected]['angle'] += 0.1*val/(2*np.pi)
            self.ROIs[self.selected]['angle'] = math.fmod(self.ROIs[self.selected]['angle'], 2*np.pi)
            self.draw()

    def get_peak_matches(self):
        patches = []
        maxv = []
        for pt, w, v in self.loc:
            if len(patches) == 0:
                patches.append((pt, w, v))
                maxv = v
            else:
                outside = len(patches)*[False]
                for j, patch in enumerate(patches):
                    x, y = patch[0][0], patch[0][1]
                    if abs(x-pt[0]) < w and abs(y-pt[1]) < w:
                        if v > maxv:
                            patches[j] = (pt, w, v)
                            maxv = v
                    elif abs(x-pt[0]) > w or abs(y-pt[1]) > w:
                        outside[j] = True
                if all(outside):
                    patches.append((pt, w, v))
                    maxv = v
        to_delete = []
        for i, (pt, w, v) in enumerate(patches):
            for id, arena in self.arenas.items():
                if arena is not None:
                    if np.sqrt( (pt[0]+w/2 - arena['x'])**2 + (pt[1]+w/2 - arena['y'])**2 ) < w:
                        to_delete.append(i)
        for i in sorted(to_delete, reverse=True):
            del patches[i]
        return patches

    def get_templates_folder(self):
        if platform.system() == 'Darwin':
            return '/Users/degoldschmidt/.pytrack/templates/{}'.format(self.__id)
        else:
            return '/home/degoldschmidt/.pytrack/templates/{}'.format(self.__id)

    def scale_by(self, val):
        if self.ROIs[self.selected] is not None:
            self.ROIs[self.selected]['radius'] += val
            if self.ROIs[self.selected]['radius'] < 1:
                self.ROIs[self.selected]['radius'] = 1
            self.draw()

    def select(self, arenaid):
        self.selected = arenaid
        self.draw()

    def get(self):
        rois = []
        for id, roi in enumerate(self.ROIs):
            if roi is None:
                pass
            else:
                rois.append({})
                rois[id]['x'] = round(roi['x'] * self.__scale,1)
                rois[id]['y'] = round(roi['y'] * self.__scale,1)
                rois[id]['radius'] = round(float(roi['radius'] * self.__scale),1)
                rois[id]['outer'] = round(float(1.25 * roi['radius'] * self.__scale),1)
                rois[id]['scale'] = round(float((roi['radius']* self.__scale)/25.),3)
                rois[id]['id'] = id
                rois[id]['spots'] = self.get_spots(id)
        return rois

if __name__ == "__main__":
    video = '/media/degoldschmidt/DATA_DENNIS_002/opto_oloop_2019-07-03T15_49_51.avi'
    #folder = '/media/degoldschmidt/DATA_DENNIS_002/backup/2018_08_01/'
    #videos = [op.join(folder, video) for video in os.listdir(folder) if video.endswith('avi') and video.startswith('cam01')]
    #print(videos)
    if platform.system() == 'Darwin':
        video = '/Users/degoldschmidt/Desktop/tracking_test_data/cam01_2017-11-24T08_26_19.avi'
    #for video in videos:
    selector = arenaROIselector(video, 'cam01', pattern=None)
    rois = selector.get()
    print(rois)
    cv2.destroyAllWindows()



import io, yaml

def read_yaml(_file):
    """ Returns a dict of a YAML-readable file '_file'. Returns None, if file is empty. """
    with open(_file, 'r') as stream:
        out = yaml.load(stream, Loader=yaml.SafeLoader)
    return out

def write_yaml(_file, _dict):
    """ Writes a given dictionary '_dict' into file '_file' in YAML format. Uses UTF8 encoding and no default flow style. """
    with io.open(_file, 'w+', encoding='utf8') as outfile:
        yaml.dump(_dict, outfile, default_flow_style=False, allow_unicode=True)



import os
import os.path as op

def get_videos(input, output, video=None):
    if 'videos' in output:
        videos = [_file for _file in output['videos']]
        dir = output['directory']
    else:
        if video is None:
            dir = input
            videos = [op.join(input, _file) for _file in os.listdir(input) if _file.endswith('avi')]
        else:
            dir = op.dirname(input)
            videos = [video]
        #videos = [op.join(input, _file) for _file in checkbox('videos', videos, msg='Select videos for tracking:')]
    return videos, dir



import cv2
import numpy as np
import os
import os.path as op
import sys
from tqdm import tqdm

from anytrack.background import get_background
from anytrack.contours import get_contours, get_contour_stats
from anytrack.files import get_videos
from anytrack.helpers import rle, nan_helper, interpolate, smooth, px, dist, in_roi
from anytrack.roiselect import arenaROIselector
from anytrack.trajectory import Trajectory
from anytrack.video import VideoCapture
from anytrack.yaml_helpers import read_yaml, write_yaml

def PrintException():
    import linecache
    exc_type, exc_obj, tb = sys.exc_info()
    f = tb.tb_frame
    lineno = tb.tb_lineno
    filename = f.f_code.co_filename
    linecache.checkcache(filename)
    line = linecache.getline(filename, lineno, f.f_globals)
    print('EXCEPTION IN ({}, LINE {} "{}"): {}'.format(filename, lineno, line.strip(), exc_obj))

def run_bg_subtraction(video, background=None, nframes=0, threshold_level=10, thresholding='dark', show=1, n_contours=4, min_size=150, max_size=250, rois=None):
    cap = VideoCapture(video,0)
    flag=False
    if nframes>0:
        nframes = nframes
    else:
        nframes = cap.len
    flytracks = [Trajectory(nframes) for i in range(n_contours)]
    for frameno in tqdm(range(nframes)):
        _,frame = cap.read()
        contours = get_contours(frame, background, threshold_level=threshold_level, thresholding=thresholding)
        contours = [cnt for cnt in contours if cv2.contourArea(cnt)>.9*min_size and cv2.contourArea(cnt)<1.1*max_size]
        output = frame.copy()
        while len(contours) > n_contours:
            if frameno > 0:
                min_dists = [np.amin([dist((flytrack.x[frameno-1],flytrack.y[frameno-1]), cv2.fitEllipse(cnt)[0]) for flytrack in flytracks]) for cnt in contours]
                print(min_dists)
                popidx = min_dists.index(max(min_dists))
            else:
                areas = [cv2.contourArea(cnt) for cnt in contours]
                print(areas)
                popidx = areas.index(max(areas))
            print(popidx)
            contours.pop(popidx)
        contour_mapping = [[in_roi(cv2.fitEllipse(cnt)[0],roi) for roi in rois].index(True) for cnt in contours] ### list of roi IDs for each contour
        if len(contours) == n_contours:
            order = contour_mapping
            for j,cnt in zip(order,contours):
                ellipse = cv2.fitEllipse(cnt)
                (x,y),(ma,mi),a = ellipse

                ax, ay = x+0.4*mi*np.cos(np.radians(a)+np.pi/2), y+0.4*mi*np.sin(np.radians(a)+np.pi/2)
                ox, oy = x-0.4*mi*np.cos(np.radians(a)+np.pi/2), y-0.4*mi*np.sin(np.radians(a)+np.pi/2)
                if frameno>0:
                    oax, oay = flytracks[j].data[frameno-1,8], flytracks[j].data[frameno-1,9]
                    oox, ooy = flytracks[j].data[frameno-1,10], flytracks[j].data[frameno-1,11]
                    if dist((ax, ay), (oax, oay))+dist((ox, oy), (oox, ooy))  > dist((ox, oy), (oax, oay))+dist((ax, ay), (oox, ooy)):
                        ax,ay,ox,oy=ox,oy,ax,ay
                px_a, px_o = frame[px(ay),px(ax),0], frame[px(oy),px(ox),0]
                flytracks[j].append(i=frameno,x=x,y=y,ma=mi,mi=ma,angle=a,avg_px=np.nanmean(frame), ax=ax, ay=ay, ox=ox, oy=oy, apx=px_a, opx=px_o)
                if show > 0:
                    cv2.circle(output,(px(x),px(y)), 1, (255,0,255),1)
                    cv2.circle(output,(px(ax),px(ay)), 2, (255,0,0),1)
                    cv2.circle(output,(px(ox),px(oy)), 2, (0,0,255),1)
                    cv2.putText(output, '{}'.format(px_a), (px(ax)+5, px(ay)+5), cv2.FONT_HERSHEY_SIMPLEX, .5, (255,0,0), 1, cv2.LINE_AA)
                    cv2.putText(output, '{}'.format(px_o), (px(ox)+5, px(oy)+5), cv2.FONT_HERSHEY_SIMPLEX, .5, (0,0,255), 1, cv2.LINE_AA)
                    pts = flytracks[j].data[:frameno,0:2].astype(np.int32)
                    pts = pts.reshape((-1,1,2))
                    cv2.polylines(output,[pts],False,(0,255,255))
                    cv2.putText(output, '{}'.format(j), (px(x)+10, px(y)+10), cv2.FONT_HERSHEY_SIMPLEX, 1, (255,0,255), 1, cv2.LINE_AA)
            for id, arena in enumerate(rois):
                ### draw arenas
                x, y, r, o  = int(arena['x']), int(arena['y']), int(arena['radius']), int(arena['outer'])
                color = (255,255,255)
                if show > 0:
                    cv2.circle(output,(x, y),r,color,1)
                    cv2.circle(output,(x, y),int(1.1*o),color,1)
                    cv2.circle(output,(x, y),1,color,-1)
                    cv2.rectangle(output, (px(x-1.5*r),px(y-1.5*r)), (px(x+1.5*r), px(y+1.5*r)), color, 1)
                    cv2.putText(output, str(id), (x+10, y-10), cv2.FONT_HERSHEY_SIMPLEX , .5, color, 1, cv2.LINE_AA)
        if show>0 or flag:
            if frameno%show==0 or flag:
                x,y,r = flytracks[0].data[frameno,0], flytracks[0].data[frameno,1], 100
                cv2.imshow("Tracking", cv2.resize(output[int(y-r):int(y+r), int(x-r):int(x+r)], (700,700)))
                if flag:
                    cv2.waitKey(0) # time to wait between frames, in mSec
                    flag = False
                else:
                    cv2.waitKey(1) # time to wait between frames, in mSec
    return flytracks

class Anytracker(object):
    def __init__(self, input=None, output=None):
        video=None
        if op.isfile(input):
            video = input
            input = op.dirname(input)
        ### create output folder inside input folder
        outfolder = op.join(input, output)
        os.makedirs(outfolder, exist_ok=True)
        outdict_file = op.join(outfolder, 'outdict.yaml')
        ### config file for multiple runs
        if op.isfile(outdict_file):
            outdict = read_yaml(outdict_file)
        else:
            outdict = {}
        #if 'folders' not in outdict:
        outdict['directory'] = input
        outdict['folders'] = {}
        outdict['folders']['background'] = op.join(outfolder, 'bg')
        outdict['folders']['image_stats'] = op.join(outfolder, 'image_stats')
        outdict['folders']['trajectories'] = op.join(outfolder, 'trajectories')
        outdict['folders']['output'] = outfolder
        for f in outdict['folders']:
            os.makedirs(outdict['folders'][f], exist_ok=True)
        ### getting videos based on input
        self.videos, basedir = get_videos(input, outdict, video=video)
        outdict['videos'] = self.videos
        self.outdict=outdict
        self.outdict_file = outdict_file
        self.background = {}

    def infer(self, num_contours):
        print('Infer contour statistics...', flush=True)
        if 'number_contours' not in self.outdict:
            self.outdict['number_contours'] = {}
            self.outdict['min_size'] = {}
            self.outdict['max_size'] = {}
        for video in tqdm(self.videos):
            if video in self.outdict['number_contours']:
                numcnts, min_size, max_size = self.outdict['number_contours'][video], self.outdict['min_size'][video], self.outdict['max_size'][video]
            else:
                numcnts, min_size, max_size = get_contour_stats(video, self.background[video], force_num_contours=num_contours)
                self.outdict['number_contours'][video] = int(round(numcnts))
                self.outdict['min_size'][video] = float(min_size)
                self.outdict['max_size'][video] = float(max_size)
            print('avg.number of contours:\t{}\nminimum area:\t{}\nmaximum area:\t{}'.format(numcnts, min_size, max_size))
        return self.outdict['number_contours'], self.outdict['min_size'], self.outdict['max_size']

    def image_stats(self, skip=30):
        print('Analyze image statistics...')
        avgs={}
        for video in tqdm(self.videos):
            if 'image_stats_files' not in self.outdict:
                self.outdict['image_stats_files'] = {}
            if video in self.outdict['image_stats_files']:                      ### load image stats
                avgs[video] = np.load(self.outdict['image_stats_files'][video])
            else:                                                               ### run image stats
                cap = VideoCapture(video, 0)
                average_intensity = np.zeros(cap.len)
                for i in tqdm(range(int(len(average_intensity)/skip))):
                    cap.set_frame(i*skip)
                    ret, image = cap.read()
                    average_intensity[i*skip] = np.nanmean(image)
                mu=np.mean(average_intensity[average_intensity>0.])
                std=np.std(average_intensity[average_intensity>0.])
                for i in tqdm(range(len(average_intensity))):
                    if i%30!=0:
                        if average_intensity[int(i/skip)*skip] > mu+std:
                            average_intensity[i] = average_intensity[int(i/skip)*skip]
                        if (int(i/skip)+1)*skip < len(average_intensity):
                            if average_intensity[(int(i/skip)+1)*skip] > mu+std:
                                average_intensity[i] = average_intensity[(int(i/skip)+1)*skip]
                binary_vec = np.zeros(average_intensity.shape, dtype=np.uint8)
                binary_vec[average_intensity>=mu+std] = 1
                outfile = op.join(self.outdict['folders']['image_stats'], '{}_stats.npy'.format(op.basename(video).split('.')[0]))
                self.outdict['image_stats_files'][video] = outfile
                np.save(outfile, binary_vec)
                avgs[video] = binary_vec
        return avgs

    def model_bg(self, baselines=None):
        print('Modelling background...', flush=True)
        if 'background_files' not in self.outdict:
            self.outdict['background_files'] = {}
        for video in tqdm(self.videos):
            if video not in self.outdict['background_files']:
                self.outdict['background_files'][video] = op.join(self.outdict['folders']['background'], '{}_bg.png'.format(op.basename(video).split('.')[0]))
            if op.isfile(self.outdict['background_files'][video]):
                self.background[video] = cv2.imread(self.outdict['background_files'][video])[:,:,0]
            else:
                if baselines is not None:
                    only_these = np.where(baselines[video]==0)[0]
                else:
                    only_these = None
                self.background[video] = get_background(video, frames=only_these)
                cv2.imwrite(self.outdict['background_files'][video], self.background[video])

    def overlay(self):
        for video in tqdm(self.videos):
            for _file in os.listdir(self.outdict['folders']['trajectories']):
                pass

    def run(self, nframes=0, threshold_level=10, thresholding='dark', use_threads=1, show=0):
        print('Run tracking...', flush=True)
        all_tracks = {}
        self.outdict['trajectory_files'] = {}
        for video in tqdm(self.videos):
            ### run tracking
            try:
                tracks = run_bg_subtraction(    video,
                                                nframes=nframes,
                                                background=self.background[video],
                                                threshold_level=threshold_level,
                                                thresholding=thresholding,
                                                n_contours=self.outdict['number_contours'][video],
                                                min_size=self.outdict['min_size'][video],
                                                max_size=self.outdict['max_size'][video],
                                                rois=self.outdict['ROIs'][video],
                                                show=show,)
                ### interpolate all signals
                for fly in tracks:
                    for i in range(fly.data.shape[1]):
                        if not all(np.isnan(fly.data[:,i])):
                            fly.data[:,i] = interpolate(fly.data[:,i])
                all_tracks[video] = tracks
            except:
                all_tracks[video] = None
                PrintException()

            for fly in tracks:
                windowlen=13
                apx = smooth(fly.data[:,12], windowlen)
                opx = smooth(fly.data[:,13], windowlen)
                diffs = opx - apx
                binary = np.sign(diffs)
                for rl, pos, state in zip(*rle(binary)):
                    if rl > 10: ### accept only valid switches longer than 10 frames
                        if state == 1:
                            fly.data[pos:pos+rl,2:4] = fly.data[pos:pos+rl,8:10]
                        elif state == -1:
                            fly.data[pos:pos+rl,2:4] = fly.data[pos:pos+rl,10:12]
                        else:
                            print(state, 'WEIRD')
                    else:
                        if state == 1:
                            fly.data[pos:pos+rl,2:4] = fly.data[pos:pos+rl,10:12]
                        elif state == -1:
                            fly.data[pos:pos+rl,2:4] = fly.data[pos:pos+rl,8:10]
                        else:
                            print(state, 'WEIRD')
                fly.data[:,6] = np.arctan2(fly.data[:,3] - fly.data[:,1], fly.data[:,2] - fly.data[:,0])

            ### save tracks
            self.outdict['trajectory_files'][video] = []
            for i,fly in enumerate(tracks):
                if fly is not None:
                    self.outdict['trajectory_files'][video].append(op.join(self.outdict['folders']['trajectories'], '{}_fly{}.csv'.format(op.basename(video).split('.')[0],i)))
                    fly.save(op.join(self.outdict['folders']['trajectories'], '{}_fly{}.csv'.format(op.basename(video).split('.')[0],i)))

        return all_tracks


    def roi_select(self, method='automated'):
        print('Initialize ROI selector...')
        self.all_rois = {}
        if 'ROIs' in self.outdict:
            self.all_rois = self.outdict['ROIs']
        for video in tqdm(self.videos):
            if video not in self.all_rois:
                selector = arenaROIselector(video, 'cam05', method=method, pattern=None)
                rois = selector.get()
                self.all_rois[video] = rois
                cv2.destroyAllWindows()
        self.outdict['ROIs'] = self.all_rois
        return self.all_rois



import argparse
from pprint import pprint
import os.path as op

from anytrack.anytracker import Anytracker
from anytrack.yaml_helpers import read_yaml, write_yaml

def main():
    ### arguments parsing
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', dest='input', action='store',
                        help='input file(s)/directory')
    parser.add_argument('-n', dest='num_contours', action='store', type=int,
                        help='fix number of contours')
    args = parser.parse_args()
    input = args.input
    num_contours = args.num_contours
    ### create AnyTrack Tracking object
    track = Anytracker(input=input, output='output_anytrack')
    ### step 1: batch run ROIfinder to get arenas (automated/supervised/manual) recommended to use 'supervised'
    track.roi_select(method='automated')
    write_yaml(track.outdict_file, track.outdict)

    ### step 2: get image statistics (average signal over time, )
    avg_pxs = track.image_stats(skip=30)
    write_yaml(track.outdict_file, track.outdict)

    ### step 3: background modelling (DONE)
    track.model_bg(baselines=avg_pxs)
    write_yaml(track.outdict_file, track.outdict)

    ### step 4: infer contour statistics (DONE)
    track.infer(num_contours=num_contours)
    write_yaml(track.outdict_file, track.outdict)

    pprint(track.outdict)
    write_yaml(track.outdict_file, track.outdict)

if __name__ == '__main__':
    main()



import os
from setuptools import setup
import distutils.util

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="anytrack",
    version="1.0.20",
    author="Dennis Goldschmidt",
    author_email="dennis.goldschmidt@neuro.fchampalimaud.org",
    description=("Simple tracking script for arbitrary object tracking in videos."),
    license="GPLv3",
    keywords=['tracking', 'data analysis', 'fly'],
    url="https://pypi.python.org/pypi/anytrack",
    packages=['anytrack'],
    python_requires='>=3.6',
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.6",
    ],
    platforms=['LinuxUbuntu', 'MacOSX-HighSierra'],
    install_requires=['numpy', 'pyyaml', 'pandas', 'scipy', 'opencv-python', 'tqdm', 'PyInquirer'],
    entry_points={
        'console_scripts': [
            'anytrack = anytrack.main:main',
            'anytrack-register = anytrack.register:main',
        ],
    },
)



from __future__ import print_function, unicode_literals
#from PyInquirer import style_from_dict, Token, prompt, Separator


def checkbox(name, options, msg=None):
    style = style_from_dict({
    Token.Separator: '#27ae60',
    Token.QuestionMark: '#00d700 bold',
    Token.Selected: '#27ae60',  # default
    Token.Pointer: '#00d700 bold',
    Token.Instruction: '',  # default
    Token.Answer: '#00d700 bold',
    Token.Question: '',
    })

    choices = [Separator('= {} ='.format(name))]
    for option in options:
        choices.append({'name': option, 'checked': True})

    if msg is None:
        msg = 'Select {}'.format(name)

    questions = [
        {
            'type': 'checkbox',
            'message': msg,
            'name': name,
            'choices': choices,
            'validate': lambda answer: 'You must choose at least one option.' \
                if len(answer) == 0 else True
        }
    ]

    return prompt(questions, style=style)[name]
